/**
  *  \file game/actions/basetech.cc
  *  \brief Starbase Action: Tech Upgrade
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/actions/basetech.h"
#include "game/undo.h"
#include "game/team.h"
#include "game/reg.h"
#include "game/actions/preconditions.h"

/*********************** GStarbaseTechUpgradeAction **********************/

/** \class GStarbaseTechUpgradeAction
    \brief Tech Upgrade rule class.

    This class encapsulates the rules for tech upgrades, and billing
    there-of. */

/** Constructor. Initializes the object and allocates resources.
    \param the_planet the planet we're talking about */
GStarbaseTechUpgradeAction::GStarbaseTechUpgradeAction(GPlanet& the_planet)
    : GStarbaseBuildAction(the_planet),
      the_planet(the_planet)
{
    // GStarbaseBuildAction will also check these, but will be constructed after us.
    // We must check that to be able to access the current tech levels.
    mustHavePlayedBase(the_planet);

    for (int i = 0; i < 4; ++i)
        new_tech[i] = min_tech[i] = the_planet.getBaseTechLevel(TTechLevel(i));

    ASSERT(getCost().isEmpty());
}

GStarbaseTechUpgradeAction::~GStarbaseTechUpgradeAction()
{ }

/** Check whether the user may raise his tech level to the specified
    value. Only checks rules, not resources. */
bool
GStarbaseTechUpgradeAction::isValidTech(int index, int new_val) const
{
    ASSERT(index >= 0 && index < 4);
    if (new_val < getMinTech(index) || new_val > getMaxTech(index))
        return false;
    return true;
}

/** Get maximum allowed tech level for this index. */
int
GStarbaseTechUpgradeAction::getMaxTech(int /*index*/) const
{
    if (allowRegistered(the_planet.getOwner())) {
        return 10;
    } else {
        return 6;
    }
}

/** Set tech level to specified value. This will not actually update
    the base, it will only register the request and update the cost
    accordingly. The actual update happens in commit().
    \return true iff request was honored (see isValid(int,int)). */
bool
GStarbaseTechUpgradeAction::setTech(int index, int new_val)
{
    if (isValidTech(index, new_val)) {
        new_tech[index] = new_val;
        sig_changed.raise();
        return true;
    } else {
        return false;
    }
}

void
GStarbaseTechUpgradeAction::perform(GStarbaseBuildExecuter& a)
{
    for (int i = 0; i < 4; ++i)
        a.setTech(TTechLevel(i), new_tech[i]);
}

/** Get error. Since we never go into an invalid state, there is no error. */
GError
GStarbaseTechUpgradeAction::getError()
{
    return GError();
}

/** Add undo information. */
void
GStarbaseTechUpgradeAction::setUndoInformation(const GUndoInformation& info)
{
    /* This does not change the cost, only the permitted range. */
    for (int i = 0; i < 4; ++i) {
        min_tech[i] = info.getMinimumTech(TTechLevel(i));
    }
}
