/**
  *  \file game/mine.cc
  *  \brief Minefield Implementation
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "game/mine.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/host.h"
#include "game/racename.h"
#include "util/misc.h"
#include "util/nls.h"

/** Create empty minefield with specified Id. */
GMinefield::GMinefield(int aid)
    : id(aid), x(0), y(0), owner(0),
      web(false), units(0), turn(0), last_action(0),
      units_prev(0), turn_prev(0),
      radius_now(0), units_now(0)
{ }

/** Postprocessing after load. */
void
GMinefield::internalCheck(const int turnNr)
{
    int32_t u = units;
    for (int i = turn; i < turnNr; ++i) {
        u = getUnitsAfterDecay(u);
    }
    units_now = u;
    radius_now = getRadiusFromUnits(units_now);
}

/** Compute mine decay for one turn. */
int32_t
GMinefield::getUnitsAfterDecay(int32_t orig_units) const
{
    int decay_rate = web ? config.WebMineDecayRate(owner) : config.MineDecayRate(owner);
    if (host.isPHost())
        /* PHost formula */
        return orig_units * (100 - decay_rate) / 100;
    else
        /* THost formula (3.22.040). Actual formula is
             ERND(orig_units - orig_units*decay_rate/100) - 1
           which should yield the same results. */
        return std::max(0L, rdivaddl(orig_units * (100-decay_rate), 100, 0) - 1);
}

/** Compute minefield radius from unit number. */
int
GMinefield::getRadiusFromUnits(int32 units)
{
    return int(std::sqrt(double(units)));
}

/** Compute successful passage rate. This is the inverse of the "hit
    rate", for a given distance. Under THost, there is only "the" hit
    rate. Under PHost, various options are arrayized, and dynamic by
    experience and speed. This function returns the value for an
    inexperienced ship owned by the current player, at warp 9 (=the
    worst possible case).

    Note that actually under THost the problem is much more
    complicated due to the interesting implementation; see
    <http://phost.de/~stefan/minehits.html>. We do not attempt to
    emulate that here.

    \param distance Distance to cover, in ly
    \param cloaked  Whether ship is cloaked
    \return Passage rate, [0.0, 1.0] */
double
GMinefield::getPassRate(double distance, bool cloaked, int player) const
{
    double rate = (isWeb()
                   ? config.WebMineHitOdds(player) * 0.01
                   : cloaked
                   ? config.MineHitOddsWhenCloakedX10(player) * 0.001
                   : config.MineHitOdds(player) * 0.01);
    if (rate <= 0) {
        /* Hit rate below zero -> pass rate is 1.0 */
        return 1.0;
    } else if (rate >= 1.0) {
        /* Hit rate is one -> pass rate is 0.0 */
        return 0.0;
    } else {
        /* Normal computation */
        return std::pow(1.0 - rate, distance);
    }
}

/** Compute units used for laying. In PHost, this is after decay; in
    THost it's before. */
int32
GMinefield::getUnitsForLaying() const
{
    if (host.isPHost()) {
        return getUnitsAfterDecay(getUnits());
    } else {
        return getUnits();
    }
}

/** Add report about this minefield.
    \param report [in] The report */
void
GMinefield::addMinefieldReport(const TMinefieldReport& report)
{
    // Is this information actually newer?
    if (report.turn < turn) {
        return;
    }

    // Is this the same field we already saw?
    const bool is_same_field = (owner == report.owner && x == report.x && y == report.y);

    // If we saw the minefield already with better reason, ignore this report.
    // For example, when laying and scooping a minefield in the same turn,
    // we get a Lay(size=X) report followed by a Sweep/Scoop(size=0) in util.dat.
    // Further Lay(size=X) reports, e.g. from messages, shall not override that.
    if (is_same_field && report.turn == turn && report.why < last_action) {
        return;
    }

    // Turn change: move previous values into archive.
    if (report.turn > turn) {
        if (is_same_field) {
            units_prev = units;
            turn_prev = turn;
        } else {
            units_prev = 0;
            turn_prev = 0;
        }
    }

    // Figure out unit count.
    if (report.units_known) {
        // Units known exactly.
        units = report.radius_or_units;
    } else {
        // Units not known exactly. Check range.
        const int32_t new_units = square(report.radius_or_units);

        /* THost uses ERND(Sqrt(units)), PHost uses Trunc(Sqrt(units)).
           Therefore, actual radius is [r,r+1) in PHost, [r-.5,r+.5] for
           THost. Hence, possible unit ranges corresponding to this
           radius are (r-.5) = r-r-.25 and (r+1). */
        const int32_t min_units = new_units - report.radius_or_units - 1;
        const int32_t max_units = square(report.radius_or_units + 1);

        if (is_same_field
            && turn == report.turn
            && units >= min_units
            && units <  max_units)
        {
            // The minefield was already seen this turn, with better information (exact unit count).
            // No change.
        } else {
            units = new_units;
        }
    }

    // Update minefield. Avoid updating the type if it is not reliably known.
    x     = report.x;
    y     = report.y;
    owner = report.owner;
    turn  = report.turn;
    if (report.type_known || !is_same_field) {
        web = report.web;
    }

    // Update cause.
    if (report.why > last_action || !is_same_field) {
        last_action = report.why;
    }
}

string_t
GMinefield::getName(Name /*which*/) const
{
    if (web)
        return format(_("Web Mine Field #%d (%s)"), id, player_racenames.getShortName(owner));
    else
        return format(_("Mine Field #%d (%s)"), id, player_racenames.getShortName(owner));
}

int
GMinefield::getId() const
{
    return id;
}

int
GMinefield::getOwner() const
{
    return owner;
}

GPoint
GMinefield::getPos() const
{
    return GPoint(x, y);
}

int
GMinefield::getRadius() const
{
    return radius_now;
}

int32_t
GMinefield::getRadiusSquared() const
{
    return units_now;
}

/***************************** GMinefieldType ****************************/

GMinefieldType::GMinefieldType(GUniverse& univ)
    : GIndexedObjectType(10000),
      univ(univ),
      all_minefields_known(),
      mines()
{ }

GMinefieldType::~GMinefieldType()
{
    done();
}

GMinefield&
GMinefieldType::getObjectByIndex(int index)
{
    return *get(index);
}

GUniverse*
GMinefieldType::getUniverseByIndex(int /*index*/)
{
    return &univ;
}

bool
GMinefieldType::isValidIndex(int index) const
{
    return mines.find(index) != mines.end();
}

void
GMinefieldType::erase(int i)
{
    minemap_t::iterator p = mines.find(i);
    if (p != mines.end()) {
        delete p->second;
        mines.erase(p);
        sig_set_changed.raise(0);
    }
}

/** Add report about a minefield.
    \param id     [in] Minefield Id
    \param report [in] The report */
void
GMinefieldType::addMinefieldReport(int id, const TMinefieldReport& report)
{
    // Get minefield, but avoid creating 0-mine fields.
    GMinefield* f;
    if (report.radius_or_units > 0) {
        minemap_t::iterator i = mines.find(id);
        if (i == mines.end()) {
            i = safeInsert(mines, id, new GMinefield(id));
        }
        f = i->second;
    } else {
        f = get(id);
    }

    // Process it.
    if (f != 0) {
        f->addMinefieldReport(report);
    }
}

/** Note that all minefields of a player are known with current data.
    This means alternatively that if we have a minefield of this player in the history,
    and did not scan it this turn, it is gone. This happens for Winplan result files
    (KORE minefields). */
void
GMinefieldType::markAllMinefieldsKnown(int player)
{
    all_minefields_known += player;
}

/** Internal check/postprocess.
    Postprocess all minefields and delete those that are gone.
    \param turnNr current turn */
void
GMinefieldType::internalCheck(const int turnNr)
{
    for (minemap_t::iterator i = mines.begin(); i != mines.end(); /* empty */) {
        GMinefield& mf = *i->second;
        mf.internalCheck(turnNr);
        if (mf.getUnits() == 0
            || (all_minefields_known.contains(mf.getOwner())
                && (mf.getTurnLastSeen() < turnNr
                    || mf.getAction() < GMinefield::mfa_Scanned)))
        {
            // Erase minefield if it's gone,
            // or should have been seen but was not seen this turn.
            minemap_t::iterator j = i;
            ++i;
            mines.erase(j);
        } else {
            ++i;
        }
    }
}

GMinefield*
GMinefieldType::get(int id) const
{
    minemap_t::const_iterator i = mines.find(id);
    if (i == mines.end()) {
        return 0;
    } else {
        return i->second;
    }
}

void
GMinefieldType::done()
{
    for (minemap_t::iterator i = mines.begin(); i != mines.end(); ++i) {
        delete i->second;
    }
}
