/**
  *  \file int/if/shipif.h
  *  \brief Interpreter Game Interface: Ships
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_IF_SHIPIF_H
#define PCC_V2_INT_IF_SHIPIF_H

#include "int/compound.h"
#include "int/context.h"
#include "int/if/objcontext.h"

class GShip;
class GCargoTransfer;
class GCargoContainer;
class GUniverse;
class IntExecutionContext;
class IntArgBlock;

IntValue*   IFShipGet(IntArgBlock& args);
IntContext* IFShipMake();
int32_t     IFShipDim(int32_t dim);

GCargoContainer* makePlanetTransfer(GUniverse& univ, int pid, GShip& to);
void doScriptTransfer(IntExecutionContext& exc, GCargoTransfer& transfer, string_t cargospec, int side, int flags);

/** Ship context.
    This class must override getObject() to achieve the same semantics as PCC 1.x
    (which seems to be the most useful semantics for the SHIP() array):
    - iteration uses ty_any_ships to iterate over everything visible
    - indexing uses ty_history_ships to allow accessing a history ship directly */
class IntShipContext : public IntObjectContext {
 public:
    IntShipContext(int sid);
    ~IntShipContext();

    // IntContext:
    int32_t   lookup(const IntNameQuery& name);
    void      set(int32_t index, IntValue* value);
    IntValue* get(int32_t index);
    IntContext* clone();
    string_t  toString(bool readable);
    void      store(IntTagNode& sv, Stream& aux);
    void      enumProperties(IntPropertyAcceptor& acceptor);

    // IntObjectContext overrider:
    GObject*  getObject();
};

/** Definition of ship properties. */
enum IntShipProperty {
    ispAuxId,
    ispAuxAmmo,
    ispAuxCount,
    ispAuxShort,
    ispAuxName,
    ispBeamId,
    ispBeamCount,
    ispBeamShort,
    ispBeamName,
    ispCargoColonists,
    ispCargoD,
    ispCargoFree,
    ispCargoM,
    ispCargoMoney,
    ispCargoN,
    ispCargoStr,
    ispCargoSupplies,
    ispCargoT,
    ispCrew,
    ispDamage,
    ispEnemyId,
    ispEngineId,
    ispEngineName,
    ispFCode,
    ispFighterBays,
    ispFighterCount,
    ispFleetId,
    ispFleetName,
    ispFleetStatus,
    ispFleet,
    ispHasFunction,
    ispHeadingAngle,
    ispHeadingName,
    ispHullSpecial,
    ispId,
    ispLevel,
    ispLocX,
    ispLocY,
    ispLoc,
    ispMarked,
    ispMass,
    ispMessages,
    ispMissionId,
    ispMissionIntercept,
    ispMissionShort,
    ispMissionTow,
    ispMissionName,
    ispMoveETA,
    ispMoveFuel,
    ispName,
    ispOrbitId,
    ispOrbitName,
    ispPlayed,
    ispRealOwner,
    ispScore,
    ispSpeedId,
    ispSpeedName,
    ispTask,
    ispTorpId,
    ispTorpCount,
    ispTorpLCount,
    ispTorpShort,
    ispTorpName,
    ispTransferShipColonists,
    ispTransferShipD,
    ispTransferShipId,
    ispTransferShipM,
    ispTransferShipN,
    ispTransferShipName,
    ispTransferShipSupplies,
    ispTransferShipT,
    ispTransferShip,
    ispTransferUnloadColonists,
    ispTransferUnloadD,
    ispTransferUnloadId,
    ispTransferUnloadM,
    ispTransferUnloadN,
    ispTransferUnloadName,
    ispTransferUnloadSupplies,
    ispTransferUnloadT,
    ispTransferUnload,
    ispTypeChar,
    ispTypeStr,
    ispWaypointDistance,
    ispWaypointDX,
    ispWaypointDY,
    ispWaypointPlanetId,
    ispWaypointX,
    ispWaypointY,
    ispWaypointName
};

enum IntShipMethod {
    ismMark,                    // 0
    ismUnmark,                  // 1
    ismSetComment,              // 2
    ismSetFCode,                // 3
    ismSetEnemy,                // 4
    ismSetSpeed,                // 5
    ismSetName,                 // 6
    ismSetMission,              // 7
    ismFixShip,                 // 8
    ismRecycleShip,             // 9
    ismSetWaypoint,             // 10
    ismCargoTransfer,           // 11
    ismCargoUnload,             // 12
    ismCargoUpload,             // 13
    ismSetFleet                 // 14
};


IntValue* getShipProperty(const GShip& sh, GUniverse& univ, IntShipProperty isp);
void      setShipProperty(GShip& sh, GUniverse& univ, IntShipProperty isp, IntValue* value);
void      callShipMethod(GShip& sh, GUniverse& univ, IntShipMethod ism, IntExecutionContext& exc, IntArgBlock& args);

#endif
