/**
  *  \file io/memstream.cc
  *  \brief Memory Stream Class
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "io/memstream.h"

using namespace std;

/****************************** MemoryStream *****************************/

/** \class MemoryStream
    \brief Stream reading/writing from a memory buffer

    This stream behaves like a regular file, but uses a
    dynamically-sized buffer to store data. The buffer grows as
    needed. There are additional functions to access the buffer. */

/** Create an empty memory stream.
    \post getPos() == 0, getSize() == 0 */
MemoryStream::MemoryStream()
    : pos(0),
      name("<memory>")
{ }

/** Create memory stream with initial contents from a string.
    \param s initial data
    \post getPos() == 0, getSize() == s.size() */
MemoryStream::MemoryStream(const string_t& s)
    : data(s.length()), pos(0)
{
    data.append(s.data(), s.size());
}

/** Create a stream with initial contents from a buffer.
    \param buf initial data. This buffer is copied. To initialize from an
               existing buffer, use the default constructor and swapBuffer().
    \post getPos() == 0, getSize() == buf.getSize() */
MemoryStream::MemoryStream(const Buffer<char>& buf)
    : data(buf.getSize()), pos(0)
{
    data.append(buf.ptr(), buf.getSize());
}

MemoryStream::~MemoryStream()
{ }

/** Get contents of stream as a string.
    \post getContentsAsString().size() == getSize()
    \post getBuffer()[i] == getContentsAsString()[i] */
string_t
MemoryStream::getContentsAsString() const
{
    return string_t(data.ptr(), data.getSize());
}

/** Swap buffer. This will exchange this stream's buffer and the
    /other/ buffer such that /other/ contains the original stream data
    and the stream contains the original data from the buffer. This
    can be used for low-overhead initialisation from a buffer, and for
    low-overhead buffer inquiry.

    \post getPos() == 0 */
void
MemoryStream::swapBuffer(Buffer<char>& other)
{
    data.swap(other);
    pos = 0;
}

std::size_t
MemoryStream::read(char* buffer, std::size_t max)
{
    if (data.getSize() - pos < max)
        max = data.getSize() - pos;
    memcpy(buffer, data.ptr() + pos, max);
    pos += max;
    return max;
}

std::size_t
MemoryStream::write(const char* buffer, std::size_t max)
{
    if (pos + max < pos)
        return 0;

    data.ensureSize(pos + max);
    memcpy(data.ptr() + pos, buffer, max);
    pos += max;
    return max;
}

int
MemoryStream::getCapabilities() const
{
    return C_READ | C_WRITE | C_SEEK;
}

void
MemoryStream::seek(fpos_t new_pos)
{
    if (new_pos >= 0 && new_pos <= data.getSize()) {
        pos = new_pos;
    } else {
        invalidFilePosition();
    }
}

MemoryStream::fpos_t
MemoryStream::getSize()
{
    return data.getSize();
}

MemoryStream::fpos_t
MemoryStream::getPos()
{
    return pos;
}

string_t
MemoryStream::getName()
{
    return name;
}

/** Set name of this stream. Sets the name returned by getName(). */
void
MemoryStream::setName(const string_t name)
{
    this->name = name;
}
